/*
 Illinois Open Source License
 
 University of Illinois/NCSA
 Open Source License
 
 Copyright © 2009,    University of Illinois.  All rights reserved.
 
 Developed by: 
 Innovative Systems Lab
 National Center for Supercomputing Applications
 http://www.ncsa.uiuc.edu/AboutUs/Directorates/ISL.html
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal with the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimers.
 
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimers in the documentation and/or other materials provided with the distribution.
 
 * Neither the names of Innovative Systems Lab and National Center for Supercomputing Applications, nor the names of its contributors may be used to endorse or promote products derived from this Software without specific prior written permission.
 
 THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
 */

#define NBINS 32
#define NJK 10

#define N1 131072
#define N2 131072
#define N1pN2 262144


double add6vals(int a1, int a2, int a3, int a4, int a5, int a6);

void cross_correlation(sram double X1[N1pN2], sram double X2[N1pN2],
                       sram double Y1[N1pN2], sram double Y2[N1pN2],
                       sram double Z1[N1pN2], sram double Z2[N1pN2],
                       sram double JK[N1pN2], sram double BINV[N1pN2],
                       int n1, int n2, int nb, int njk, int doSelf)
{
    int i, i1, j, k, jk1, jk2, offset, indx1, indx2, nb1, nb2, i_end, j_start, val;
    double x1, y1, z1, x2, y2, z2, x3, y3, z3, dot1, dot2;

    // NBINS
    double binb00, binb01, binb02, binb03, binb04, binb05;
    double binb06, binb07, binb08, binb09, binb10, binb11;
    double binb12, binb13, binb14, binb15, binb16, binb17;
    double binb18, binb19, binb20, binb21, binb22, binb23;
    double binb24, binb25, binb26, binb27, binb28, binb29;
    double binb30;

    // NJK
    int binvA1_00[NBINS], binvA2_00[NBINS], binvA3_00[NBINS];
    int binvA1_01[NBINS], binvA2_01[NBINS], binvA3_01[NBINS];
    int binvA1_02[NBINS], binvA2_02[NBINS], binvA3_02[NBINS];
    int binvA1_03[NBINS], binvA2_03[NBINS], binvA3_03[NBINS];
    int binvA1_04[NBINS], binvA2_04[NBINS], binvA3_04[NBINS];
    int binvA1_05[NBINS], binvA2_05[NBINS], binvA3_05[NBINS];
    int binvA1_06[NBINS], binvA2_06[NBINS], binvA3_06[NBINS];
    int binvA1_07[NBINS], binvA2_07[NBINS], binvA3_07[NBINS];
    int binvA1_08[NBINS], binvA2_08[NBINS], binvA3_08[NBINS];
    int binvA1_09[NBINS], binvA2_09[NBINS], binvA3_09[NBINS];
    int binvA1_10[NBINS], binvA2_10[NBINS], binvA3_10[NBINS];

    int binvB1_00[NBINS], binvB2_00[NBINS], binvB3_00[NBINS];
    int binvB1_01[NBINS], binvB2_01[NBINS], binvB3_01[NBINS];
    int binvB1_02[NBINS], binvB2_02[NBINS], binvB3_02[NBINS];
    int binvB1_03[NBINS], binvB2_03[NBINS], binvB3_03[NBINS];
    int binvB1_04[NBINS], binvB2_04[NBINS], binvB3_04[NBINS];
    int binvB1_05[NBINS], binvB2_05[NBINS], binvB3_05[NBINS];
    int binvB1_06[NBINS], binvB2_06[NBINS], binvB3_06[NBINS];
    int binvB1_07[NBINS], binvB2_07[NBINS], binvB3_07[NBINS];
    int binvB1_08[NBINS], binvB2_08[NBINS], binvB3_08[NBINS];
    int binvB1_09[NBINS], binvB2_09[NBINS], binvB3_09[NBINS];
    int binvB1_10[NBINS], binvB2_10[NBINS], binvB3_10[NBINS];

    int bin_bank;
    int valA0, valA1, valA2, valA3, valA4, valA5, valA6, valA7, valA8, valA9, valA10;
    int valB0, valB1, valB2, valB3, valB4, valB5, valB6, valB7, valB8, valB9, valB10;

    // make a local copy of bin boundaries
    nb1 = nb + 1;
    nb2 = nb + 2;

    // copy bin boundaries
    for (i = 0; i < nb1; i++)
    {
        binb00 = binb01;
        binb01 = binb02;
        binb02 = binb03;
        binb03 = binb04;
        binb04 = binb05;
        binb05 = binb06;
        binb06 = binb07;
        binb07 = binb08;
        binb08 = binb09;
        binb09 = binb10;
        binb10 = binb11;
        binb11 = binb12;
        binb12 = binb13;
        binb13 = binb14;
        binb14 = binb15;
        binb15 = binb16;
        binb16 = binb17;
        binb17 = binb18;
        binb18 = binb19;
        binb19 = binb20;
        binb20 = binb21;
        binb21 = binb22;
        binb22 = binb23;
        binb23 = binb24;
        binb24 = binb25;
        binb25 = binb26;
        binb26 = binb27;
        binb27 = binb28;
        binb28 = binb29;
        binb29 = binb30;
        binb30 = BINV[n1 + i];
    }

    for (i = 0; i < nb2; i++)
    {
        binvA1_00[i] = 0; binvA2_00[i] = 0; binvA3_00[i] = 0;
        binvA1_01[i] = 0; binvA2_01[i] = 0; binvA3_01[i] = 0;
        binvA1_02[i] = 0; binvA2_02[i] = 0; binvA3_02[i] = 0;
        binvA1_03[i] = 0; binvA2_03[i] = 0; binvA3_03[i] = 0;
        binvA1_04[i] = 0; binvA2_04[i] = 0; binvA3_04[i] = 0;
        binvA1_05[i] = 0; binvA2_05[i] = 0; binvA3_05[i] = 0;
        binvA1_06[i] = 0; binvA2_06[i] = 0; binvA3_06[i] = 0;
        binvA1_07[i] = 0; binvA2_07[i] = 0; binvA3_07[i] = 0;
        binvA1_08[i] = 0; binvA2_08[i] = 0; binvA3_08[i] = 0;
        binvA1_09[i] = 0; binvA2_09[i] = 0; binvA3_09[i] = 0;
        binvA1_10[i] = 0; binvA2_10[i] = 0; binvA3_10[i] = 0;

        binvB1_00[i] = 0; binvB2_00[i] = 0; binvB3_00[i] = 0;
        binvB1_01[i] = 0; binvB2_01[i] = 0; binvB3_01[i] = 0;
        binvB1_02[i] = 0; binvB2_02[i] = 0; binvB3_02[i] = 0;
        binvB1_03[i] = 0; binvB2_03[i] = 0; binvB3_03[i] = 0;
        binvB1_04[i] = 0; binvB2_04[i] = 0; binvB3_04[i] = 0;
        binvB1_05[i] = 0; binvB2_05[i] = 0; binvB3_05[i] = 0;
        binvB1_06[i] = 0; binvB2_06[i] = 0; binvB3_06[i] = 0;
        binvB1_07[i] = 0; binvB2_07[i] = 0; binvB3_07[i] = 0;
        binvB1_08[i] = 0; binvB2_08[i] = 0; binvB3_08[i] = 0;
        binvB1_09[i] = 0; binvB2_09[i] = 0; binvB3_09[i] = 0;
        binvB1_10[i] = 0; binvB2_10[i] = 0; binvB3_10[i] = 0;
    }

    if (doSelf) i_end = n1-1;
    else i_end = n1;

    // loop thru one dataset
    for (i = 0; i < i_end; i += 2)
    {
        x1 = X1[i];
        y1 = Y1[i];
        z1 = Z1[i];
        jk1 = (int)JK[i];

        i1 = i + 1;
        x3 = X2[i1];
        y3 = Y2[i1];
        z3 = Z2[i1];
        jk2 = (int)JK[i1];

        if (doSelf) j_start = i1;
        else j_start = 0;

        // loop thru second dataset
        for (j = j_start; j < n2; j++)
        {
            if (doSelf) offset = j;
            else offset = n1 + j;

            x2 = X2[offset];
            y2 = Y2[offset];
            z2 = Z2[offset];

            // dot product
            dot1 = x1 * x2 + y1 * y2 + z1 * z2;
            dot2 = x3 * x2 + y3 * y2 + z3 * z2;

            // run binary search
            #include "bsearch1.c"
            #include "bsearch2.c"

            bin_bank = j % 3;

            // update bin values
            #include "binupdate1.c"
            #include "binupdate2.c"
        }
    }

    // collapce partial results and copy back to SRAM
    for (i = 0; i < nb2; i++)
    {
               BINV[i] = add6vals(binvA1_00[i], binvA2_00[i], binvA3_00[i], binvB1_00[i], binvB2_00[i], binvB3_00[i]);
           BINV[nb2+i] = add6vals(binvA1_01[i], binvA2_01[i], binvA3_01[i], binvB1_01[i], binvB2_01[i], binvB3_01[i]);
         BINV[2*nb2+i] = add6vals(binvA1_02[i], binvA2_02[i], binvA3_02[i], binvB1_02[i], binvB2_02[i], binvB3_02[i]);
         BINV[3*nb2+i] = add6vals(binvA1_03[i], binvA2_03[i], binvA3_03[i], binvB1_03[i], binvB2_03[i], binvB3_03[i]);
         BINV[4*nb2+i] = add6vals(binvA1_04[i], binvA2_04[i], binvA3_04[i], binvB1_04[i], binvB2_04[i], binvB3_04[i]);
         BINV[5*nb2+i] = add6vals(binvA1_05[i], binvA2_05[i], binvA3_05[i], binvB1_05[i], binvB2_05[i], binvB3_05[i]);
         BINV[6*nb2+i] = add6vals(binvA1_06[i], binvA2_06[i], binvA3_06[i], binvB1_06[i], binvB2_06[i], binvB3_06[i]);
         BINV[7*nb2+i] = add6vals(binvA1_07[i], binvA2_07[i], binvA3_07[i], binvB1_07[i], binvB2_07[i], binvB3_07[i]);
         BINV[8*nb2+i] = add6vals(binvA1_08[i], binvA2_08[i], binvA3_08[i], binvB1_08[i], binvB2_08[i], binvB3_08[i]);
         BINV[9*nb2+i] = add6vals(binvA1_09[i], binvA2_09[i], binvA3_09[i], binvB1_09[i], binvB2_09[i], binvB3_09[i]);
        BINV[10*nb2+i] = add6vals(binvA1_10[i], binvA2_10[i], binvA3_10[i], binvB1_10[i], binvB2_10[i], binvB3_10[i]);
    }

}


double add6vals(int a1, int a2, int a3, int a4, int a5, int a6)
{
    int res;
    res = ((a1 + a2) + (a3 + a4) + (a5 + a6));
    return (double)res;
}